from BubbleGun.graph_io import read_gfa, write_gfa, write_chains
from BubbleGun.compact_graph import compact_graph
from BubbleGun.connected_components import all_components
from BubbleGun.bfs import bfs
import sys
import os


class Graph:
    """
    Graph object containing the important information about the graph
    """

    __slots__ = ['nodes', 'b_chains', 'bubbles', 'compacted']

    def __init__(self, graph_file=None, low_memory=False):
        if graph_file is not None:
            if not os.path.exists(graph_file):
                print("graph file {} does not exist".format(graph_file))
                sys.exit()
            # loading nodes from file
            self.nodes = read_gfa(gfa_file_path=graph_file, low_memory=low_memory)
        else:
            self.nodes = dict()
        # elif graph_file.endswith(".vg"):
        #     self.nodes = read_vg(vg_file_path=graph_file, k=k, modified=modified, coverage=coverage)

        self.b_chains = set()  # list of BubbleChain objects
        self.bubbles = dict()
        self.compacted = False
        # self.k = 1

    def __len__(self):
        """
        overloading the length function
        """

        return len(self.nodes)

    def __str__(self):
        """
        overloading the string function for printing
        """

        return "The graph has {} Nodes and {} chains".format(
            len(self.nodes), len(self.b_chains))

    # def __repr__(self):
    #     """
    #     overloading the string function for printing
    #     """
    #     return "The graph has {} Nodes and {} bubble and {} chains".format(
    #         len(self.nodes), len(self.bubbles), len(self.b_chains))

    def add_chain(self, chain):
        """
        adds a bubble chain to the graph
        """
        if len(chain.sorted) == 0:
            chain.find_ends()
            chain.sort()
            if len(chain.ends) != 2:  # circular chains or other weird stuff
                nodes_set = set(chain.list_chain())
                self.write_graph(set_of_nodes=nodes_set, append=True,
                                 output_file="circular_and_other_problematic_chains.gfa", optional_info=False)

            else:
                # self.b_chains[chain._BubbleChain__key()] = chain
                if chain not in self.b_chains:
                    self.b_chains.add(chain)

    def total_seq_length(self):
        """
        returns total sequence length
        """
        total = 0
        for n in self.nodes.values():
            total += n.seq_len
        return total
        # total = 0
        # counted_overlap = set()
        # for n in self.nodes.values():
        #
        #     total += n.seq_len
        #     if n.id not in counted_overlap:
        #         counted_overlap.add(n.id)
        #
        #         for nn in n.end:
        #             counted_overlap.add(nn[0])
        #             total -= nn[2]
        #         for nn in n.start:
        #             counted_overlap.add(nn[0])
        #             total -= nn[2]
        #
        # return total

    def longest_chain_bubble(self):
        """
        returns the longest bubble chain bubble wise
        In case there are more than one longest chain it
        returns the first one found
        """

        lengths_list = [len(x) for x in self.b_chains]
        m = max(lengths_list)
        m_idx = lengths_list.index(m)
        return list(self.b_chains)[m_idx]

    def longest_chain_seq(self):
        """
        returns the longest bubble chain sequence wise
        In case there are more than on longest chain it
        returns the first one found
        """

        lengths_list = [x.length_seq() for x in self.b_chains]
        m = max(lengths_list)
        m_idx = lengths_list.index(m)
        # returning only one chain that is the max, there could be a tie
        return list(self.b_chains)[m_idx]

    def nodes_in_chains(self):
        """
        returns the set of all nodes in bubble chains
        """
        # pdb.set_trace()
        all_nodes = []
        for chain in self.b_chains:
            all_nodes += chain.list_chain()
        return set(all_nodes)

    def seq_in_chains(self):
        """
        returns how much sequence their are in the bubble chains
        """
        # total_length = 0
        # for chain in self.b_chains:
        #     chain_length = 0
        #     counted_overlap = set()
        #     chain_nodes = set(chain.list_chain(ids=False))
        #     for n in chain_nodes:
        #         chain_length += n.seq_len
        #         if n.id not in counted_overlap:
        #             counted_overlap.add(n.id)
        #             for nn in n.end:
        #                 if nn in chain_nodes:
        #                     counted_overlap.add(nn[0])
        #                     chain_length -= nn[2]
        #             for nn in n.start:
        #                 if nn in chain_nodes:
        #                     counted_overlap.add(nn[0])
        #                     chain_length -= nn[2]
        #     total_length += chain_length
        # return total_length
        s_in_c = 0
        for chain in self.b_chains:
            # if chain not in self.child_parent:
            s_in_c += chain.length_seq()
        return s_in_c

    def chain_cov_node(self):
        """
        returns the percentage the nodes in chains covered
        """
        # pdb.set_trace()
        n_in_c = self.nodes_in_chains()
        return float((len(n_in_c) * 100) / len(self.nodes))

    def chain_cov_seq(self):
        """
        returns the percentage the sequences in chains covered
        """
        chains_nodes = set()
        for chain in self.b_chains:
            for n in chain.list_chain():
                chains_nodes.add(n)
        total_seq = 0
        for n in chains_nodes:
            total_seq += self.nodes[n].seq_len
        return (total_seq*100) / float(self.total_seq_length())

    def num_single_bubbles(self):
        """
        return the number of chains that has only one bubble
        """

        nsb = 0
        for chain in self.b_chains:
            if len(chain) == 1:
                nsb += 1
        return nsb

    def reset_visited(self):
        """
        resets all nodes.visited to false
        """

        for n in self.nodes.values():
            n.visited = False

    def bubble_number(self):
        """
        returns a list of 3 counters, bubbles, superbubbles and insertions
        """

        counter = [0, 0, 0]
        for chain in self.b_chains:
            for b in chain.bubbles:
                if b.is_simple():
                    counter[0] += 1
                elif b.is_super():
                    counter[1] += 1
                elif b.is_insertion():
                    counter[2] += 1

        return counter

    # def find_chains(self, only_simple=False):
    #     """
    #     calls the find bubbles and chains algorithms
    #     then adds the objects to the graph
    #     """
    #
    #     find_bubble_chains(self, only_simple)

    def remove_node(self, n_id):
        """
        remove a node and its corresponding edges
        """
        starts = [x for x in self.nodes[n_id].start]
        for n_start in starts:
            overlap = n_start[2]
            if n_start[1] == 1:
                self.nodes[n_start[0]].end.remove((n_id, 0, overlap))
            else:
                self.nodes[n_start[0]].start.remove((n_id, 0, overlap))

        ends = [x for x in self.nodes[n_id].end]
        for n_end in ends:
            overlap = n_end[2]
            if n_end[1] == 1:
                self.nodes[n_end[0]].end.remove((n_id, 1, overlap))
            else:
                self.nodes[n_end[0]].start.remove((n_id, 1, overlap))

        del self.nodes[n_id]

    def remove_lonely_nodes(self):
        """
        remove singular nodes with no neighbors
        """

        nodes_to_remove = [n.id for n in self.nodes.values() if len(n.neighbors()) == 0]

        for i in nodes_to_remove:
            self.remove_node(i)

    def compact(self):
        """
        compact the unipaths in the graph
        and turns the graph into a compacted one
        """

        # if self.k == 0:
        #     logging.warning("if this is De Bruijn Graph"
        #                     " and you did not specify the k value"
        #                     " the compacting might not be correct, as overlaps"
        #                     " needs to be removed")
        compact_graph(self)
        self.compacted = True

    def write_graph(self, set_of_nodes=None,
                    output_file="output_graph.gfa",
                    append=False, optional_info=True):
        """writes a graph file as GFA

        list_of_nodes can be a list of node ids to write
        ignore_nodes is a list of node ids to not write out
        if append is set to true then output file should be an existing
        graph file to append to
        modified to output a modified graph file
        """
        if not output_file.endswith(".gfa"):
            output_file += ".gfa"

        write_gfa(self, set_of_nodes=set_of_nodes, output_file=output_file,
                  append=append, optional_info=optional_info)

    def write_b_chains(self, output="bubble_chains.gfa"):
        """writes bubble gains to a GFA file

        :params output: output file path
        """
        if not output.endswith(".gfa"):
            output += ".gfa"

        if self.compacted:
            write_chains(self, output_file=output, optional_info=False)
        else:
            write_chains(self, output_file=output, optional_info=True)

    def biggest_comp(self):
        """
        returns a list of sets for each component and the node ids in that
        component.
        """
        all_comp = all_components(self)
        lengths = [len(i) for i in all_comp]
        return all_comp[lengths.index(max(lengths))]

    def bfs(self, start, size):
        """
        Returns a neighborhood of size given around start node

        :param start: starting node for the BFS search
        :param size: size of the neighborhood to return
        """

        neighborhood = bfs(self, start, size)
        return neighborhood

    def fill_bubble_info(self):
        # chains_to_remove = set()
        b_counter = 0
        chain_num = 0
        for chain in self.b_chains:
            chain_num += 1  # To start from 1

            for bubble in chain.sorted:
                b_counter += 1
                if bubble.is_simple():
                    # randomly assigning which branch is zero and which is 1
                    # simple bubble only has 2 nodes inside, so 0 and 1
                    for allele, node in enumerate(bubble.inside):
                        # check if it hasn't been processed yet
                        # it might be a nested bubble inside an SB
                        # so don't need to change the tags then
                        # SB takes precedence
                        if node.which_chain == 0:
                            node.which_allele = allele
                            node.which_chain = chain_num
                            node.which_b = b_counter

                    # assigning the ends of the bubble to which chain
                    if bubble.source.which_chain == 0:
                        bubble.source.which_chain = chain_num

                    if bubble.sink.which_chain == 0:
                        bubble.sink.which_chain = chain_num

                else:  # superbubbles or insertions
                    for node in bubble.list_bubble():
                        # inside the SB it gets reset
                        self.nodes[node].which_b = 0
                        self.nodes[node].which_allele = -1
                        self.nodes[node].which_sb = b_counter
                        self.nodes[node].which_chain = chain_num



import json
import os
import sys
import pdb


def write_certain_chains(json_file, graph, chain_ids, output_file):
    """
    Output certain chains based on their ids
    """

    set_of_nodes = set()
    chain_ids = set(chain_ids)
    if not os.path.exists(json_file):
        print("File {} does not exists, make sure tha path is correct".format(json_file))
        sys.exit(1)

    with open(json_file, "r") as in_file:
        for line in in_file:
            chains = json.loads(line)
            for c_id in chain_ids:
                if c_id not in chains:
                    continue

                wanted_chain = chains[c_id]
                set_of_nodes = set_of_nodes.union(set(wanted_chain['ends']))
                for bubble in wanted_chain['bubbles']:
                    set_of_nodes = set_of_nodes.union(set(bubble['ends']))
                    set_of_nodes = set_of_nodes.union(set(bubble['inside']))

    graph.write_graph(set_of_nodes=set_of_nodes, output_file=output_file)



class Bubble:
    """
    Bubble object that has the important information about a bubble
    """
    __slots__ = ['source', 'sink', 'inside', 'key', 'id', 'parent_chain', 'parent_sb', 'chain_id']

    def __init__(self, source, sink, inside):
        """
        Initialize the bubble object
        """
        self.source = source
        self.sink = sink
        self.inside = inside
        self.key = self.__key()
        self.id = 0
        self.parent_chain = 0
        self.parent_sb = 0
        self.chain_id = 0
        # self.visited = False  # I need this when finding chains

    def __len__(self):
        """
        overloading the length function
        """
        return len(self.inside) + 2

    def __key(self):
        source = str(self.source.id)
        sink = str(self.sink.id)
        if source > sink:
            return (source, sink)
        return (sink, source)

    # I need to hash the bubbles because I can find the same bubble
    # twice as I'm coming from different directions
    # this way I can avoid adding it twice to the list of bubbles
    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return self.key == other.key

    def __ne__(self, other):
        return not self.__eq__(other)

    def list_bubble(self):
        """"
        returns a list of node ids that make up the bubble
        """
        node_list = [self.source.id, self.sink.id]
        for n in self.inside:
            node_list.append(n.id)
        return node_list

    def length_node(self):
        """
        returns how many nodes in the bubble including source and sink
        """
        return len(self.inside) + 2  # +2 for source and sink

    def length_seq(self):
        """
        returns the total sequence in the bubble
        """
        total_seq = self.source.seq_len + self.sink.seq_len
        for n in self.inside:
            total_seq += n.seq_len
        return total_seq

    def is_simple(self):
        """
        returns true if it's a simple bubble

        I need to make sure it's just two disjoint paths, because the algorithm can still report a bubble with 2 nodes inside but having different directed edges.

        note to self: to flip direction I can always do abs(direction - 1) if 1 it becomes 0, if 0 it becomes 1

        - First check I need is to make sure the source and sink are not connected directly with each other
        - check that middle nodes are only connected to source and sink
        """
        # checking that the nodes inside only connected to two nodes each (supposedly the source and sink)
        if len(self.inside) != 2:
            return False
        if not {1} == set([len(self.inside[0].start), len(self.inside[0].end), len(self.inside[1].start), len(self.inside[1].end)]):
            return False
        # check that the nodes inside are only connected to source and sink
        neighbors1 = self.inside[0].neighbors()
        neighbors2 = self.inside[1].neighbors()
        neighbors1.sort()
        neighbors2.sort()
        if not neighbors1 == neighbors2:
            return False

        # make sure source and sink are not connected
        if self.source.id in self.sink.neighbors() or self.sink.id in self.source.neighbors():
            return False

        return True

        # if len(self.inside) == 2:
        #     return True
        # return False

    def is_insertion(self):
        """
        returns true if it's an insertion
        i.e. bubble with one branch
        """
        if len(self.inside) != 1:
            return False

        if not {1} == set([len(self.inside[0].start), len(self.inside[0].end)]):
            return False

        neighbors = self.inside[0].neighbors()
        neighbors.sort
        tmp = [self.source.id, self.sink.id]
        tmp.sort()
        if tmp != neighbors:
            return False
        return True
        # if len(self.inside) == 1:
        #     return True
        # return False

    def is_super(self):
        """
        returns true if it's a superbubble
        """
        if not self.is_simple():
            if not self.is_insertion():
                return True
        return False
        # if len(self.inside) > 2:
        #     return True
        # return False

    def set_as_visited(self):
        """
        sets all the nodes of the bubble as visited
        """
        self.source.visited = True
        self.sink.visited = True
        for n in self.inside:
            n.visited = True



#!/usr/bin/env python3
import sys
import argparse
import logging
from BubbleGun.bfs import bfs
from BubbleGun.Graph import Graph
from BubbleGun.bubbles_fasta import write_bubbles
from BubbleGun.find_bubbles import find_bubbles
from BubbleGun.fasta_chains import output_chains_fasta
from BubbleGun.json_out import json_out
from BubbleGun.output_certain_chains import write_certain_chains
from BubbleGun.connect_bubbles import connect_bubbles
from BubbleGun.find_parents import find_parents

parser = argparse.ArgumentParser(description='Find Bubble Chains.', add_help=True)
subparsers = parser.add_subparsers(help='Available subcommands', dest="subcommands")

parser._positionals.title = 'Subcommands'
parser._optionals.title = 'Global Arguments'

########################## general commands ###############################

parser.add_argument("-g", "--in_graph", metavar="GRAPH_PATH", dest="in_graph",
                    default=None, type=str, help="graph file path (GFA or VG)")

parser.add_argument("--log_file", dest="log_file", type=str, default="log.log",
                        help="The name/path of the log file. Default: log.log")

parser.add_argument("--log", dest="log_level", type=str, default="DEBUG",
                    help="The logging level [DEBUG, INFO, WARNING, ERROR, CRITICAL]")

########################## Bubble chains options ###############################
bubble_parser = subparsers.add_parser('bchains', help='Command for detecting bubble chains')

bubble_parser.add_argument("--bubble_json", dest="out_json", default=None, type=str,
                           help="Outputs Bubbles, Superbubbles, and Chains as a JSON file")

bubble_parser.add_argument("--only_simple", action="store_true",
                           help="If used then only simple bubbles are detected")

bubble_parser.add_argument("--only_super", action="store_true",
                           help="If used then only simple bubbles are detected")

bubble_parser.add_argument("--save_memory", action="store_true", dest="low_memory",
                           help="Identifies bubble chain with less memory. No statistics outputted")

bubble_parser.add_argument("--chains_gfa", dest="chains_gfa", default=None, type=str,
                           help="Output only bubble chains as a GFA file")

bubble_parser.add_argument("--fasta", dest="out_fasta", metavar="FASTA",
                           type=str, default=None,
                           help="Outputs the bubble branches as fasta file (doesn't work with memory saving)")

bubble_parser.add_argument("--out_haplos", dest="out_haplos", action="store_true",
                           help="output randomly two haplotypes for each chain (doesn't work with memory saving)")

########################## Compact graph ###############################
compact_parser = subparsers.add_parser('compact', help='Command for compacting graphs')

compact_parser.add_argument(dest="compacted", metavar="PATH_COMPACTED",
                            default=None, type=str, help="Compacted graph output path")

########################## Biggest component ###############################
biggest_comp_parser = subparsers.add_parser('biggestcomp', help='Command for separating biggest component')

biggest_comp_parser.add_argument(dest="biggest_comp", metavar="PATH_BIG_COMP",
                                 default=None, type=str, help="Biggest component output path")

########################## BFS commands ###############################
bfs_parser = subparsers.add_parser('bfs', help='Command for separating neighborhood')

bfs_parser.add_argument("--start", dest="starting_nodes", metavar="START_NODES", type=str, nargs="+",
                        default=None, help="Give the starting node(s) for neighborhood extraction")

bfs_parser.add_argument("--neighborhood_size", dest="bfs_len", metavar="SIZE", default=None,
                        type=int, help="With -s --start option, size of neighborhood to extract")

bfs_parser.add_argument("--output_neighborhood", dest="output_neighborhood", metavar="OUTPUT",
                        type=str, default=None, help="Output neighborhood file")

# ########################## Digest gam ###############################
# gam_parser = subparsers.add_parser('gamdigest', help='Command for digesting a gam file')
#
# gam_parser.add_argument("--json_file", dest="json_file", metavar="JSON_FILE",
#                         type=str, default=None, help="The JSON file wtih bubble chains information")
#
# gam_parser.add_argument("--alignment_file", dest="gam_file", metavar="GAM",
#                         type=str, default=None, help="Take GAM file and output pickled dict")
#
# gam_parser.add_argument("--min_cutoff", dest="min_cutoff", type=int, default=None,
#                         help="The minimum cutoff of a mapping length.")
#
# gam_parser.add_argument("--out_dict", dest="pickle_out",
#                         type=str, default=None,
#                         help="A pickled dictionary output path. contains read_id:[nodes]")

########################## output chain ###############################
output_chain = subparsers.add_parser('chainout', help='Outputs certain chain(s) given by their id as a GFA file')
output_chain.add_argument("--json_file", dest="json_file", metavar="JSON_FILE",
                          type=str, default=None, help="The JSON file wtih bubble chains information")

output_chain.add_argument("--chain_ids", dest="chain_ids", metavar="CHAIN_IDS", type=str, nargs="+",
                          default=None, help="Give the chain Id(s) to be outputted")

output_chain.add_argument("--output_chain", dest="output_chain", metavar="OUTPUT",
                          type=str, default=None, help="Output path for the chains chosen")

args = parser.parse_args()

# log_file = "log_" + str(time.clock_gettime(1)).split(".")[0] + ".log"
log_file = args.log_file

logging.basicConfig(filename=log_file, filemode='w',
                    format='[%(asctime)s] %(message)s',
                    level=getattr(logging, args.log_level.upper()))

logging.info(" ".join(["argument given:"] + sys.argv))


def main():
    if len(sys.argv) == 1:
        print("You did not provide any arguments\n"
              "Try to use -h or --help for help\n")
        sys.exit()

    if args.subcommands is None:
        print("Please provide a subcommand after the global commands")
        sys.exit(1)

    if args.in_graph is None:
        print("You need to provide an input graph with -g")
        sys.exit(1)
    # if args.in_graph is None:
    #     if args.subcommands != "gamdigest":
    #         print("you didn't give an input graph file")
    #         parser.print_help()
    #         sys.exit(0)

    ####################### chainout
    if args.subcommands == "chainout":
        if args.json_file is not None:
            if args.chain_ids is not None:
                if args.output_chain is not None:

                    logging.info("Reading Graph...")
                    graph = Graph(args.in_graph)

                    logging.info("Outputting chains chosen...")
                    write_certain_chains(args.json_file, graph, args.chain_ids, args.output_chain)

                else:
                    print("You did not provide the output file path")
                    sys.exit(1)
            else:
                print("You did not provide a chain id(s) to be outputted")
                sys.exit(1)
        else:
            print("You did not provide the JSON file with the chain information")
            sys.exit(1)

    # ####################### gamdigest
    # if args.subcommands == "gamdigest":
    #     if args.gam_file is not None:
    #         if args.pickle_out is not None:
    #             if args.min_cutoff is not None:
    #                 logging.info("reading gam file and building dict")
    #                 all_reads = digest_gam(args.in_graph, args.gam_file, args.min_cutoff, args.pickle_out)
    #                 logging.info("finished successfully")
    #                 # sys.exit()
    #             else:
    #                 print("You did not provide the minimum cutoff. Maybe something like 3 times the k-mer length")
    #                 sys.exit(1)
    #         else:
    #             print("You did not provide a path for the pickled dictionary output")
    #             sys.exit(1)
    #     else:
    #         print("Please provide the path to the gam file")
    #         sys.exit(1)

    ####################### compact graph
    if args.subcommands == "compact":
        if args.compacted is not None:
            logging.info("Reading Graph...")
            graph = Graph(args.in_graph)

            logging.info("Compacting Graph...")
            graph.compact()
            logging.info("Writing Compacted Graph...")
            graph.write_graph(output_file=args.compacted, optional_info=False)
            logging.info("Done...")
        else:
            print("Please provide the path for the output compacted graph")
            sys.exit(1)

    ####################### biggest component
    if args.subcommands == "biggestcomp":
        if args.biggest_comp is not None:
            logging.info("Reading Graph...")

            graph = Graph(args.in_graph)

            logging.info("Finding Biggest Component...")
            biggest_comp = graph.biggest_comp()
            logging.info("Writing Biggset Component...")
            graph.write_graph(output_file=args.biggest_comp, set_of_nodes=biggest_comp)
            logging.info("Done...")
        else:
            print("Please provide the path for the biggest component graph")
            sys.exit(1)

    ####################### Bubbles
    if args.subcommands == "bchains":
        # output_file = args.out_bubbles
        if (args.low_memory and (args.out_fasta is not None)) or (args.low_memory and (args.chains_gfa is not None)):
            print("You cannot combine memory saving with --fasta or --chains_gfa")
            sys.exit(0)

        if args.out_haplos:
            if args.low_memory:
                print("You cannot have save memory and output haplos")
                sys.exit()
            if not args.only_simple:
                print("If you want 2 haplotyhpes out, then you need to choose --only_simple")
                sys.exit()
            if args.only_super:
                print("You cannot have out haplotypes and both only_simple and only_super, only works with only_simple")
                sys.exit()

        logging.info("Reading Graph...")
        graph = Graph(args.in_graph, low_memory=args.low_memory)

        logging.info("Finding bubbles...")

        find_bubbles(graph, only_simple=args.only_simple, only_super=args.only_super)

        logging.info("Connecting bubbles...")
        # connecting individual bubbles into chains
        connect_bubbles(graph)

        # add information related to nested bubbles
        logging.info("Finding nested information...")
        find_parents(graph)

        logging.info("Done finding chains...")
        b_numbers = graph.bubble_number()
        print("The number of Simple Bubbles is {}\n"
              "The number of Superbubbles is {}\n"
              "The number of insertions is {}".format(b_numbers[0], b_numbers[1],
                                                      b_numbers[2]))

        if (0,0,0) != (b_numbers[0], b_numbers[1],b_numbers[2]):

            if not args.low_memory:
                print("Sequence coverage of the bubble chains is {}%".format(graph.chain_cov_seq()))
                print("Node coverage of the bubble chains is {}%".format(graph.chain_cov_node()))
                print("The longest chain seq-wise has {} bp".format(graph.longest_chain_seq().length_seq()))
                print("The longest chain bubble_wise has {} bubbles".format(len(graph.longest_chain_bubble())))

            if args.out_json is not None:
                logging.info("Outputting bubble chains gfa...")
                json_out(graph, args.out_json)

            if args.chains_gfa is not None:
                if args.low_memory:
                    print("You cannot have low memory and output GFA file of chains")
                    sys.exit()

                logging.info("Outputting bubble chains gfa...")
                graph.write_b_chains(output=args.chains_gfa)

            if args.out_fasta is not None:
                if args.low_memory:
                    print("You cannot have low memory and output fasta")
                    sys.exit()

                logging.info("Outputting each bubble branch...")
                write_bubbles(graph, args.out_fasta)

            if args.out_haplos:
                logging.info("Outputting two random haplotypes of each bubble chain...")
                output_chains_fasta(graph)

    ####################### BFS
    if args.subcommands == "bfs":
        if args.starting_nodes is not None:
            if args.bfs_len is not None:
                if args.output_neighborhood is not None:

                    logging.info("Reading Graph...")
                    graph = Graph(args.in_graph)

                    for n in args.starting_nodes:
                        logging.info("extracting neighborhood around node {}".format(n))
                        set_of_nodes = bfs(graph, n, args.bfs_len)
                        if not graph.compacted:
                            graph.write_graph(set_of_nodes=set_of_nodes,
                                              output_file=args.output_neighborhood, append=True, optional_info=True)
                        else:
                            graph.write_graph(set_of_nodes=set_of_nodes,
                                              output_file=args.output_neighborhood, append=True, optional_info=False)
                        logging.info("finished successfully...")
                else:
                    print("You need to give an output file name --output_neighborhood")
                    sys.exit()
            else:
                print("You did not give the neighborhood size")
                sys.exit(1)
        else:
            print("You did not give the starting node(s)")
        logging.info("Done...")

if __name__ == "__main__":
    main()



import sys
from BubbleGun.Bubble import Bubble


"""
Pseudo code for finding Super- and Simple-Bubbles
for node s in nodes:
    push s in S
    for direction in s:  # which direction to look for children
        while len(s) > 0:
            pop arbitrary v from S
            v.visited = True
            if v has no children in direction:
                break
            for u in v's children in direction:
                u_parents = parents from where v connects to u
                if u == s:
                    break
                u.seen = True
                if all of u_parents are visited:
                    push u into S
            if len(S) == 1 (contains t):
                and no edge between t and s:
                    return t
                else
                    break
"""


def find_sb_alg(graph, s, direction, only_simple=False, only_super=False):
    """
    takes the graph and a start node s and add a bubble to the chain 
    if one is found if s was the source
    """
    # I tuples of node ids and the direction
    seen = set()
    visited = set()
    nodes_inside = []
    seen.add((s.id, direction))
    # seen.add(s.id)
    S = {(s, direction)}
    while len(S) > 0:

        v = S.pop()
        v = (v[0], v[1])
        visited.add(v[0].id)

        nodes_inside.append(v[0])

        # it's visited so it's not in the seen list anymore
        seen.remove((v[0].id, v[1]))

        # from which direction we are going we take those children
        if v[1] == 0:
            children = v[0].start
        else:
            children = v[0].end

        if len(children) == 0:
            # it's a tip
            break

        for u in children:
            # check where we entered to get children from the other side
            if u[1] == 0:
                u_child_direction = 1
                u_parents = [x[0] for x in graph.nodes[u[0]].start]

            else:
                u_child_direction = 0
                u_parents = [x[0] for x in graph.nodes[u[0]].end]

            if u[0] == s.id:
                # we are in a loop
                S = set()  # so I exit the outer loop too
                break

            # adding child to seen
            # seen.add(u[0])
            if u[1] == 0:
                seen.add((u[0], 1))
            else:
                seen.add((u[0], 0))
            # if all u_parents are visited then we push it into S
            if all(graph.nodes[i].id in visited for i in u_parents):
                S.add((graph.nodes[u[0]], u_child_direction))

        # checking if we finished
        if (len(S) == 1) and (len(seen) == 1):
            t = S.pop()
            nodes_inside.append(t[0])

            if len(nodes_inside) == 2:
                # it's an empty bubble
                # this shouldn't happen if the graph is compacted
                break

            # t[0].visited = True

            # because I'm looking in both directions I end up finding each
            # bubble twice, so I can hash the id of source and sink
            # and see if I've already found it or not
            nodes_inside.remove(s)
            nodes_inside.remove(t[0])
            bubble = Bubble(source=s, sink=t[0], inside=nodes_inside)

            if only_simple:
                if bubble.is_simple():
                    return bubble
            elif only_super:
                # If it's a bubble and it's not simple or an insertion
                # then it's marked as super
                if not bubble.is_simple():
                    if not bubble.is_insertion():
                        return bubble
            else:
                return bubble

            # if we already found the same bubble from another direction
            # I don't think I need to check because I'm finding bubbles sequentially
            # but I will investigate this later

            # if bubble not in chain:
            #     if only_simple:
            #         if bubble.is_simple():
            #             chain.add_bubble(bubble)
            #         else:
            #             # todo add end of chain here and test
            #             break
            #     elif only_super:
            #         if bubble.is_super():
            #             chain.add_bubble(bubble)
            #
            #     else:
            #         chain.add_bubble(bubble)
            #     # calling the function again on sink
            #     # and continuing the search to chain bubbles
            #     find_sb_alg(graph, t[0], t[1], chain, only_simple)
            #
            # else:
            #     # todo add end of chain here and test
            #     break
    return None


def children_of_children(graph, children):
    """
    returns the children of the children in a list
    """

    # I need to check that each branch has the same neighbors
    # to make sure it's a simple bubble (the neighbors should be 
    # the source and sink)
    c_of_c1 = graph.nodes[children[0]].neighbors()
    c_of_c2 = graph.nodes[children[1]].neighbors()

    if c_of_c1 == c_of_c2:
        return c_of_c1
    return None


def find_b_alg(graph, s, direction, chain):
    """
    takes a graph and a start node s and adds a bubble object to the chain
    if one is found, otherwise it returns None
    This is fast for finding simple bubbles only
    """
    children = s.children(direction)

    # The source doesn't have 2 children the algorithm aborts
    if len(children) != 2:
        return None
    elif all(graph.nodes[i].visited for i in children):
        return None
    elif (len(children) == 2) and (s.id in children):
        return None

    c_of_c = children_of_children(graph, children)

    if c_of_c is None:
        return None

    if c_of_c:
        c_of_c.remove(s.id)

    if len(c_of_c) == 1:

        bubble = Bubble(source=s,
                        sink=graph.nodes[c_of_c[0]],
                        inside=[graph.nodes[children[0]],
                                graph.nodes[children[1]]])

        if bubble not in chain:
            bubble.set_as_visited()
            chain.add_bubble(bubble)

        # find next bubble if exists in chain
        for d in [1, 0]:
            find_b_alg(graph, graph.nodes[c_of_c[0]], d, chain)


def find_bubbles(graph, only_simple=False, only_super=False, list_of_nodes=None):
    """
    main function for finding bubbles
    Takes a graph and fills in the bubble chains
    """
    # counter = 0

    # print("[{}] Sorting nodes\n".format(current_time()))
    # sorting nodes by length
    # sorted_nodes = sorted(graph.nodes, key=lambda x: graph.nodes[x].seq_len, reverse=True)
    # find_b_alg(graph, graph.nodes[167], 0, chains)
    # pdb.set_trace()

    if only_simple and only_super:
        print("You can't mix both only_super and only_simple, choose one or not add these arguments to detect both")
        sys.exit(1)

    if list_of_nodes is None:
        list_of_nodes = graph.nodes.values()

    for n in list_of_nodes:
        # print(n.id)
        # chain = BubbleChain()
        # counter += 1
        # if (counter % 1000000) == 0:
        #     print("[{}] {} nodes have been processed".format(current_time(), counter))
        if not n.visited:

            for d in [0, 1]:  # looking in both direction for each node
                # find_b_alg(graph, n, d, chain)
                bubble = find_sb_alg(graph, n, d, only_simple, only_super)

                if bubble is not None:
                    graph.bubbles[bubble.key] = bubble
                    # graph.bubbles.add(bubble)
            # if len(chain) != 0:
            #     # chain.find_ends()
            #     graph.add_chain(chain)

                # for n in chain.list_chain(ids=False):
                #     n.visited = True



from time import time
from datetime import datetime


def current_time():
    return datetime.fromtimestamp(time()).strftime('%Y-%m-%d %H:%M:%S')


def reverse_complement(dna):
    complement = {'A': 'T', 'C': 'G', 'G': 'C', 'T': 'A', 'N':'N'}
    return ''.join([complement[base] for base in dna[::-1]])


def calculate_n50(graph):
    list_of_lengths = []
    for n in graph.nodes.values():
        list_of_lengths.append(n.seq_len)
    list_of_lengths = sorted(list_of_lengths)
    new_list = []
    for x in list_of_lengths:
        new_list += [x]*x
    if len(new_list) % 2 == 0:
        medianpos = int(len(new_list)/2)
        return float(new_list[medianpos] + new_list[medianpos-1])/2
    else:
        medianpos = int(len(new_list)/2)
        return new_list[medianpos]


def bfs(graph, start_node, size):
    """
    Returns a neighborhood of size given around start node

    :param graph: A graph object from class Graph
    :param start_node: starting node for the BFS search
    :param size: size of the neighborhood to return
    """

    queue = []
    neighborhood = set()
    visited = set()

    queue.append(start_node)
    visited.add(start_node)
    neighborhood.add(start_node)

    neighbors = graph.nodes[start_node].neighbors()

    if len(neighbors) == 0:  # start node was a lonely node
        return neighborhood

    # break when the subgraph is longer than the size wanted
    # or the component the start node in was smaller than the size
    while (len(neighborhood) <= size) and len(queue) > 0:
        start = queue.pop()

        if start not in neighborhood:
            neighborhood.add(start)

        visited.add(start)
        neighbors = graph.nodes[start].neighbors()
        for n in neighbors:
            if n not in visited:
                queue.append(n)

    return neighborhood


from BubbleGun.functions import reverse_complement
import pdb

"""
Compact a graph

It's an updated version of the old recursive way to do compactness
This one uses a while loop instead of nested functions
Nested functions took too much memory when there were many nodes to compact
This doesn't take any extra memory than what's already stored.
"""


def merge_end(graph, n):
    # print("in merge END with n {} and neighbor {}".format(n, graph.nodes[n].end[0]))
    # print("and n's START are {}".format(graph.nodes[n].start))
    nodes = graph.nodes
    # k = graph.k

    if n != nodes[n].end[0][0]:  # no self loops
        neighbor = nodes[n].end[0]
        # checking if the neighbor is connected at start (so we have + + edge)
        # and that it only have one node from the start which is n
        if (neighbor[1] == 0) and (len(nodes[neighbor[0]].start) == 1):
            # the ends of n becomes the ends of neighbor
            # I am copying the connections of neighbors on the opposite side
            # and removing neighbor node
            new_ends = [x for x in nodes[neighbor[0]].end]
            nodes[n].end += [x for x in nodes[neighbor[0]].end]
            # the sequence and seq_len gets updated
            nodes[n].seq += nodes[neighbor[0]].seq[neighbor[2]:]
            # nodes[n].seq += nodes[neighbor[0]].seq[k - 1:]
            # adding the coverages and taking the mean
            # nodes[n].coverage = (nodes[n].coverage + nodes[neighbor[0]].coverage)/2

            nodes[n].seq_len = len(nodes[n].seq)
            # nodes[neighbor[0]] = None
            # print(f"Removing node {neighbor[0]}")
            graph.remove_node(neighbor[0])
            # I think I can remove the neighbor node here with
            # Here I need to check the new neighbors at end, 
            # and remove the merged nod
            # and add n to them

            # I needed new_ends because I can't iterate and change
            # nodes[n].end at the same time
            for nn in new_ends:
                overlap = nn[2]
                if nn[1] == 0:
                    # nodes[nn[0]].start.remove((neighbor[0], 1))
                    nodes[nn[0]].start.append((n, 1, overlap))
                elif nn[1] == 1:
                    # the if else here needed in case of there was a self 
                    # loop on the end side of neighbor
                    # the self loops is added to the merged node
                    if nn[0] != neighbor[0]:
                        # nodes[nn[0]].end.remove((neighbor[0], 1))
                        nodes[nn[0]].end.append((n, 1, overlap))
                    else:
                        try:
                            nodes[n].end.remove((neighbor[0], 1, overlap))
                            nodes[n].end.append((n, 1, overlap))
                        except:
                            pdb.set_trace()

            # if I reached here then I successfully compacted a node
            return True
            # if len(nodes[n].end) == 1:
            #     merge_end(graph, n)

        elif (neighbor[1] == 1) and (len(nodes[neighbor[0]].end) == 1):
            # the ends of n becomes the start of neighbor (because it's flipped)
            # and the sequence and seq_len get updated
            # nodes[n].end = copy.deepcopy(nodes[neighbor[0]].start)
            # I think I can remove the neighbor node here with
            new_ends = [x for x in nodes[neighbor[0]].start]
            nodes[n].end += [x for x in nodes[neighbor[0]].start]

            reverse = reverse_complement(nodes[neighbor[0]].seq)
            nodes[n].seq += reverse[neighbor[2]:]
            # nodes[n].seq += reverse[k - 1:]
            # nodes[n].coverage = (nodes[n].coverage + nodes[neighbor[0]].coverage)/2
            nodes[n].seq_len = len(nodes[n].seq)
            # nodes[neighbor[0]] = None
            # if neighbor[0] == 3827:
            #     pdb.set_trace()
            graph.remove_node(neighbor[0])
            # Here I need to check the new neighbors at end of n, 
            # and remove the merged node
            # and add n to them
            for nn in new_ends:
                overlap = nn[2]
                if nn[1] == 0:
                    # the if else here needed in case of there was a 
                    # self loop on the end side of neighbor
                    # the self loops is added to the merged node
                    if nn[0] != neighbor[0]:
                        # nodes[nn[0]].start.remove((neighbor[0], 0))
                        nodes[nn[0]].start.append((n, 1, overlap))
                    else:
                        try:
                            nodes[n].end.remove((neighbor[0], 0, overlap))
                            nodes[n].end.append((n, 1, overlap))
                        except:
                            pdb.set_trace()

                elif nn[1] == 1:
                    # nodes[nn[0]].end.remove((neighbor[0], 0))
                    nodes[nn[0]].end.append((n, 1, overlap))

            return True
            # if len(nodes[n].end) == 1:
            #     merge_end(graph, n)
    # if I'm here that mean none of the True returns were reached
    return False


def merge_start(graph, n):
    # print("in merge START with n {} and neighbor {}".format(n, graph.nodes[n].start[0]))
    # print("and n's END are {}".format(graph.nodes[n].end))

    nodes = graph.nodes
    # k = graph.k
    if n != nodes[n].start[0][0]:  # no self loop
        neighbor = nodes[n].start[0]
        # checking if the neighbor is connected at start (so we have - + edge)
        # and that it only have one node from the start which is n
        if (neighbor[1] == 0) and (len(nodes[neighbor[0]].start) == 1):
            # the start of n becomes the ends of neighbor
            # and the sequence and seq_len get updated
            # nodes[n].start = copy.deepcopy(nodes[neighbor[0]].end)
            starts = [x for x in nodes[neighbor[0]].end]
            # print(f"adding these new starts to start")
            nodes[n].start += [x for x in nodes[neighbor[0]].end]

            reverse = reverse_complement(nodes[neighbor[0]].seq)
            # nodes[n].seq = nodes[neighbor[0]].seq[:nodes[neighbor[0]].seq_len - (k - 1)] + nodes[n].seq

            nodes[n].seq = reverse[:len(reverse) - neighbor[2]] + nodes[n].seq
            nodes[n].seq_len = len(nodes[n].seq)
            # nodes[neighbor[0]] = None
            # if neighbor[0] == 3827:
            #     pdb.set_trace()
            graph.remove_node(neighbor[0])
            # Here I need to check the new neighbors at end, and remove the merged node
            # and add n to them
            for nn in starts:
                overlap = nn[2]
                # We are connected to it from start
                if nn[1] == 0:
                    # nodes[nn[0]].start.remove((neighbor[0], 1))
                    nodes[nn[0]].start.append((n, 0, overlap))

                elif nn[1] == 1:
                    if nn[0] != neighbor[0]:
                        # nodes[nn[0]].end.remove((neighbor[0], 1))
                        nodes[nn[0]].end.append((n, 0, overlap))
                    else:
                        try:
                            nodes[n].start.remove((neighbor[0], 1, overlap))
                            nodes[n].start.append((n, 0, overlap))
                        except:
                            pdb.set_trace()

            return True
            # if len(nodes[n].start) == 1:
            #     merge_start(graph, n)

        elif (neighbor[1] == 1) and (len(nodes[neighbor[0]].end) == 1):
            # the start of n becomes the ends of neighbor
            # and the sequence and seq_len get updated
            # nodes[n].start = copy.deepcopy(nodes[neighbor[0]].end)
            starts = [x for x in nodes[neighbor[0]].start]
            nodes[n].start += [x for x in nodes[neighbor[0]].start]

            reverse = reverse_complement(nodes[neighbor[0]].seq)
            nodes[n].seq = reverse[:len(reverse) - neighbor[2]] + nodes[n].seq
            # nodes[n].seq = reverse[:len(reverse) - (k - 1)] + nodes[n].seq
            nodes[n].seq_len = len(nodes[n].seq)
            # nodes[neighbor[0]] = Non
            # if neighbor[0] == 3827:
            #     pdb.set_trace()
            graph.remove_node(neighbor[0])
            # Here I need to check the new neighbors at end, and remove the merged node
            # and n to them
            for nn in starts:
                overlap = nn[2]
                # We are connected to new neighbor from start
                if nn[1] == 0:
                    if nn[0] != neighbor[0]:
                        nodes[nn[0]].start.append((n, 0, overlap))
                    else:
                        try:
                            nodes[n].start.remove((neighbor[0], 0, overlap))
                            nodes[n].start.append((n, 0, overlap))
                        except:
                            pdb.set_trace()
                elif nn[1] == 1:
                    nodes[nn[0]].end.append((n, 0, overlap))
            return True
    return False


def compact_graph(graph):
    # need to have a list of nodes to iterate through
    # can't iterate and change the same dictionary
    list_of_nodes = list(graph.nodes.keys())
    for n in list_of_nodes:
        if n in graph.nodes:
            # print("at node {}".format(n))
            # print("begin at node {}".format(n))
            # # checking it's not a node that already got merged in the while loop
            #     if graph.nodes[n] is not None:
            #         # checking if it has one neighbor and it's not a self loop
            # print("I am at node {}".format(n))
            while True:
                # pdb.set_trace()
                if len(graph.nodes[n].end) == 1:
                    # print("first while at node {}".format(n))
                    if not merge_end(graph, n):
                        break
                else:
                    break

            while True:
                # pdb.set_trace()
                # print("second while at node {}".format(n))
                if len(graph.nodes[n].start) == 1:
                    # print("at node {}".format(n))
                    if not merge_start(graph, n):
                        break
                else:
                    break



from BubbleGun.functions import reverse_complement


def next_direction(node1, node2, direction):
    """
    returns the next direction to look for a neighbor

    :param node1: node object
    :param node2: another node object
    :return: 0 or 1 for next direction
    """
    if direction == 0:
        for n in node1.start:
            if node2.id == n[0]:
                return 1 - n[1], n[2]  # flipping 0 and 1

    elif direction == 1:
        for n in node1.end:
            if node2.id == n[0]:
                return 1 - n[1], n[2]  # flipping 0 and 1


def path_checker(graph, path):
    """
    checks if there's an actual path connecting the ordered nodes in path

    :param path: list of node ids, ordered according to the path
    :param graph: Graph object
    :return: Boolean
    """

    if len(path) <= 1:
        return False

    # I don't know where to look first
    if graph.nodes[path[0]].in_direction(path[1], 0):
        direction = 0
    elif graph.nodes[path[0]].in_direction(path[1], 1):
        direction = 1
    else:
        # first two nodes are not connected
        return False

    # todo somewhere here is wrong
    for i in range(len(path) - 1):
        current_node = graph.nodes[path[i]]
        next_node = graph.nodes[path[i+1]]

        if current_node.in_direction(next_node.id, direction):
            direction, overlap = next_direction(current_node, next_node, direction)
        else:
            return False

    return True


def sequence_extractor(graph, path):
    """
    returns the sequence of the path

    :param graph: a graph object
    :param path: a list of nodes ordered according to the path
    :return: sequence of the path
    """

    # check if path exists
    if len(path) == 1:
        return graph.nodes[path[0]].seq

    elif not path_checker(graph, path):
        return ""

    if graph.nodes[path[0]].in_direction(graph.nodes[path[1]].id, 0):
        direction = 0
        sequence = reverse_complement(graph.nodes[path[0]].seq)
    elif graph.nodes[path[0]].in_direction(graph.nodes[path[1]].id, 1):
        direction = 1
        sequence = graph.nodes[path[0]].seq

    for i in range(len(path) - 1):
        current_node = graph.nodes[path[i]]
        next_node = graph.nodes[path[i+1]]
        if current_node.in_direction(next_node.id, direction):
            direction, overlap = next_direction(current_node, next_node, direction)
            # if next direction is one this means current node connects to
            # next node from 0 so I don't need to take the reverse complement
            # Otherwise I need to
            if direction == 1:
                sequence += next_node.seq[overlap:]
            else:
                sequence += reverse_complement(next_node.seq)[overlap:]

    return sequence



from BubbleGun.path_extractor import sequence_extractor


def write_bubbles(graph, output_path):
    out_file = open(output_path, "w")

    for chain in graph.b_chains:
        for bubble in chain.bubbles:
            if not bubble.is_simple():
                continue

            first_path = [bubble.source.id, bubble.inside[0].id, bubble.sink.id]
            seq1 = sequence_extractor(graph, first_path)
            second_path = [bubble.source.id, bubble.inside[1].id, bubble.sink.id]
            seq2 = sequence_extractor(graph, second_path)

            read_name1 = ">bubble_" + str(bubble.id) + "_allele_1" + "_chain_" + str(chain.id)
            read_name2 = ">bubble_" + str(bubble.id) + "_allele_2" + "_chain_" + str(chain.id)

            out_file.write(read_name1 + "\n")
            out_file.write(seq1 + "\n")
            out_file.write(read_name2 + "\n")
            out_file.write(seq2 + "\n")

    out_file.close()



import logging
import pdb
from BubbleGun.BubbleChain import BubbleChain


def connect_bubbles(graph):
    """
    takes a graph object and connects the individual bubbles into chains and add them back to graph object
    """

    ids_to_bubbles = dict()
    for b in graph.bubbles.values():
        if b.source.id not in ids_to_bubbles:
            ids_to_bubbles[b.source.id] = [b]
        else:
            ids_to_bubbles[b.source.id].append(b)

        if b.sink.id not in ids_to_bubbles:
            ids_to_bubbles[b.sink.id] = [b]
        else:
            ids_to_bubbles[b.sink.id].append(b)

    starting_nodes = [x for x in ids_to_bubbles if len(ids_to_bubbles[x]) == 1]

    # a source or a sink can be part of two bubbles at most
    # so lists in ids_to_bubbles can only be 1 or 2 in length
    for n in starting_nodes:
        chain = BubbleChain()
        if ids_to_bubbles[n]:
            current_b = ids_to_bubbles[n][0]
        else:
            continue
        current_n = n
        while True:

            if current_n == current_b.source.id:
                next_n = current_b.sink.id
            else:
                next_n = current_b.source.id

            ids_to_bubbles[next_n].remove(current_b)

            chain.add_bubble(current_b)

            if len(ids_to_bubbles[next_n]) < 1:
                break

            current_b = ids_to_bubbles[next_n][0]  # this has only one bubble
            current_n = next_n

        if len(chain) != 0:
            chain.find_ends()
            graph.add_chain(chain)  # get sorted and ends found when added to graph

    # filling bubbles and chains ids
    b_counter = 1
    chain_counter = 1
    for chain in graph.b_chains:
        chain.id = chain_counter

        for b in chain.sorted:
            b.id = b_counter
            b.chain_id = chain.id
            b_counter += 1
        chain_counter += 1

    # pdb.set_trace()



def find_component(graph, start_node):
    """
    Find component in the graph starting from given node

    :param graph: is a graph object from class Graph
    :param start_node: The start node of BFS search.
    :return: a list of node ids for the component
    """
    queue = []
    cc = set()

    # visited = set()
    queue.append(start_node)
    graph.nodes[start_node].visited = True
    # visited.add(start_node)
    neighbors = graph.nodes[start_node].neighbors()

    if len(neighbors) == 0:
        cc.add(start_node)
        return cc

    while len(queue) > 0:
        start = queue.pop()
        if start not in cc:
            cc.add(start)
        else:
            continue

        # visited.add(start)
        graph.nodes[start].visited = True
        neighbors = graph.nodes[start].neighbors()
        for n in neighbors:
            if not graph.nodes[n].visited:
                queue.append(n)

    return cc


def all_components(graph):
    """
    find all connected components in the graph

    :params graph: is a graph object from class Graph
    :return: list of list of components
    """

    graph.reset_visited()  # turns all nodes to False
    connected_comp = []
    # visited = set()
    for n in graph.nodes:
        # if n not in visited:
        if not graph.nodes[n].visited:
            connected_comp.append(find_component(graph, n))
            # visited = visited.union(connected_comp[-1])
    return connected_comp



import logging
from BubbleGun.find_bubbles import find_sb_alg
import pdb

"""
Sort all Sbs from biggest to smallest
then starting from biggest, find nested bubbles, assign parent to be the sb I started with
repeat for each sb

In theory, the bubbles in one chain should have the same parent sb, which should solve the problem
"""


def list_in_list(list1, list2):
    for n in list1:
        if n not in list2:
            return False
    return True


def find_parents(graph):
    all_sbs = [x for x in graph.bubbles.values() if x.is_super()]

    # todo: check this with a bigger example
    all_sbs = sorted(all_sbs, key=lambda x: len(x.inside))
    all_sbs.reverse()

    """
    Best way I can do this is by turning graph.bubbles into a dictionary
    Then it's easier to change the values with bubbles that have a parent id 
    """
    for sb in all_sbs:
        list_of_nodes = sb.inside
        inside_bubbles = set()
        for n in list_of_nodes:
            for d in [0, 1]:
                bubble = find_sb_alg(graph, n, d)
                if bubble is not None:
                    # bubble.parent_sb = sb
                    if bubble.key in graph.bubbles:
                        graph.bubbles[bubble.key].parent_sb = sb.id
                        graph.bubbles[bubble.key].parent_chain = sb.chain_id
                    # print(id(bubble))
                    # pdb.set_trace()

    for chain in graph.b_chains:
        for b in chain.bubbles:
            if b.parent_sb != 0:
                chain.parent_sb = b.parent_sb
                chain.parent_chain = b.parent_chain
                break



from collections import Counter


class BubbleChain:
    """
    BubbleChain object which is a set of bubble objects
    """
    __slots__ = ['bubbles', 'sorted', 'ends', 'key', 'id', 'parent_chain', 'parent_sb']

    def __init__(self):
        """
        initialize the BubbleChain as a set of bubble
        """
        self.bubbles = set()
        self.sorted = []  # sorted bubble pointers
        self.ends = []  # node ids of the chain ends
        self.id = 0
        self.parent_sb = 0
        self.parent_chain = 0
        # self.key = self.__hash__()

    def __key(self):
        """
        calculated the key of the bubble chain
        """
        if self.ends[0] > self.ends[1]:
            return self.ends[0], self.ends[1]
        return self.ends[1], self.ends[0]

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        return self.__key() == other.__key()

    def __ne__(self, other):
        return not self.__eq__(other)

    def __len__(self):
        """
        overloading the length function
        """
        return len(self.bubbles)

    def __contains__(self, item):
        """
        Overloading membership operator
        """
        return item in self.bubbles

    def add_bubble(self, bubble):
        """
        adds a bubble object to the chain
        """
        self.bubbles.add(bubble)

    def list_chain(self, ids=True):
        """
        return all nodes in the chain as a list of node objects
        """
        c_list = []
        for b in self.bubbles:
            c_list += [b.source, b.sink] + b.inside
        if ids:
            return list(set([x.id for x in c_list]))

        return list(set(c_list))  # set to remove redundant sources and sinks

    def length_node(self):
        """
        returns how many nodes there are in the chain
        """
        return len(self.list_chain())

    def length_seq(self):
        """
        returns sequence length covered by the chain
        """
        # total_seq = 0
        # counted_overlaps = set()
        # for n in self.list_chain(ids=False):
        #     total_seq += n.seq_len
        #     if n.id not in counted_overlaps:
        #         for nn in n.end:
        #             counted_overlaps.add(nn[0])
        #             total_seq -= nn[2]
        #         for nn in n.start:
        #             counted_overlaps.add(nn[0])
        #             total_seq -= nn[2]
        total_seq = 0
        for n in self.list_chain(ids=False):
            total_seq += n.seq_len
        return total_seq

    def find_ends(self):
        """
        Find the ends of a chain

        todo maybe add the ends while constructing the chain in find_sb_alg
        """
        self.ends = [k for k, v in Counter([b.source.id for b in self.bubbles] + [b.sink.id for b in self.bubbles]).items() if v == 1]

    def sort(self):
        """
        sorts the bubbles in the chain
        """
        # finding ends
        all_ends = dict()
        for b in self.bubbles:
            source = str(b.source.id)
            sink = str(b.sink.id)
            if source > sink:
                all_ends[(source, sink)] = b
            else:
                all_ends[(sink, source)] = b

            # looks at all the sources, sinks of the bubbles in the chain
            # only the ends of the chain should have a count of 1
            # the rest are counted twice as two adjacent bubbles will share
            # the same node as sink for one and source for the other
        # if len(self.ends) == 0:
        #     self.ends = [k for k, v in Counter([n for sublist in all_ends.keys() for n in sublist]).items() if v == 1]
        # try:
        #     assert len(self.ends) == 2
        # except AssertionError:
        #     pdb.set_trace()
        # I couldn't sort the set of bubbles (overload a "bigger than" function
        # in Bubble to use for the python sort function) because the sorting
        # here is based on the chain and the ends of the chain
        # I don't know before hand which bubble is "bigger" or "smaller" than
        # the other bubble until I have the complete chain
        # for example, if I start traversing from the middle of the chain
        # looking "left" or "right" I still don't know who's is "bigger"
        # until I finish the whole chain.
        # pdb.set_trace()
        start = self.ends[0]  # randomly choosing one end of the chain as start
        all_keys = list(all_ends.keys())
        while len(self.sorted) < len(self.bubbles):
            for idx, key in enumerate(all_keys):
                if start in key:
                    rm_key = idx
                    start = key[1 - key.index(start)]
                    self.sorted.append(all_ends[key])
                    break
            del all_keys[rm_key]



import json
import pdb


def json_out(graph, output):
    # For each chain I output each bubble's ends and insides
    # First I'll give each bubble an id so the children chain can have a parent id
    # now the parent is just a pointer to a sb object and a chain object
    # b_counter = 1
    # chain_counter = 1
    # for chain in graph.b_chains.values():
    #     chain.id = chain_counter
    #     for b in chain.sorted:
    #         b.id = b_counter
    #         b_counter += 1
    #     chain_counter += 1

    # pdb.set_trace()
    output_f = open(output, "w")
    # output_f.write("[\n")
    json_out = dict()
    # # I am writing the nested chains first
    # for child_chain_key, parent_info in graph.child_parent.items():
    #     chain_line = dict()
    #
    #     chain_line['chain_id'] = child_chain_key.id
    #     chain_line['ends'] = child_chain_key.ends
    #     chain_line['bubbles'] = []
    #     chain_line['parent_chain'] = parent_info[0].id
    #     chain_line['parent_sb'] = parent_info[1].id
    #     for bubble in child_chain_key.sorted:
    #         line = dict()
    #         if bubble.is_simple():
    #             line['type'] = 'simple'
    #         elif bubble.is_super():
    #             line['type'] = 'super'
    #         else:
    #             line['type'] = 'insertion'
    #
    #         line['id'] = bubble.id
    #         line["ends"] = [bubble.source.id, bubble.sink.id]
    #         line['inside'] = [x.id for x in bubble.inside]
    #         chain_line['bubbles'].append(line)
    #
    #     json_out[child_chain_key.id] = chain_line
    #     # output_f.write(json.dumps(chain_line) + ",\n")
    # not nested chains
    for chain in graph.b_chains:
        # The chains that are not nested
        chain_line = dict()
        chain_line['chain_id'] = chain.id
        chain_line['ends'] = chain.ends
        chain_line['bubbles'] = []

        for bubble in chain.sorted:
            line = dict()

            if bubble.is_simple():
                line['type'] = 'simple'
                # line['id'] = bubble.inside[0].which_b

            elif bubble.is_super():
                line['type'] = 'super'
                # line['id'] = bubble.inside[0].which_sb

            else:
                line['type'] = 'insertion'
                # line['id'] = bubble.inside[0].which_s

            line['id'] = bubble.id
            line["ends"] = [bubble.source.id, bubble.sink.id]
            line['inside'] = [x.id for x in bubble.inside]

            chain_line['bubbles'].append(line)

        if bubble.parent_sb != 0:
            chain_line['parent_sb'] = bubble.parent_sb
            chain_line['parent_chain'] = bubble.parent_chain

        json_out[chain.id] = chain_line
            # output_f.write(json.dumps(chain_line) + ",\n")
    output_f.write(json.dumps(json_out))
    # output_f.write("]")
    output_f.close()



import sys


class Node:
    __slots__ = ['id', 'seq', 'seq_len', 'start', 'end', 'visited', 'optional_info']

    def __init__(self, identifier):
        self.id = identifier  # size is between 28 and 32 bytes
        self.seq = ""
        self.seq_len = 0  # mostly 28 bytes
        self.start = []  # 96 bytes for 4 neighbors
        self.end = []  # 96 bytes
        self.visited = False  # 28 bytes (used for bubble and superbubble detection)
        self.optional_info = ""
        # bubble related information
        # self.which_chain = 0
        # self.which_sb = 0
        # self.which_b = 0
        # self.which_allele = -1
        # self.coverage = 0

    def __sizeof__(self):
        size = self.id.__sizeof__() + self.seq_len.__sizeof__() + self.visited.__sizeof__()

        if len(self.start) == 0:
            size += self.start.__sizeof__()
        else:
            for i in range(len(self.start)):
                size += sys.getsizeof(self.start[i])

        if len(self.end) == 0:
            size += self.end.__sizeof__()
        else:
            for i in range(len(self.end)):
                size += sys.getsizeof(self.end[i])

        return size

    def neighbors(self):
        """
        Returns all adjacent nodes to self
        """

        neighbors = [x[0] for x in self.start] + [x[0] for x in self.end]
        return sorted(neighbors)

    def in_direction(self, node, direction):
        """
        returns true if node is a neighbor in that direction
        """

        if direction == 0:
            if node in [x[0] for x in self.start]:
                return True
            return False
        else:
            if node in [x[0] for x in self.end]:
                return True
            return False

    def children(self, direction):
        """
        returns the children of a node in given direction
        """

        if direction == 0:
            return [x[0] for x in self.start]
        elif direction == 1:
            return [x[0] for x in self.end]
        else:
            raise Exception("Trying to access a wrong direction in node {}".format(self.id))



from collections import deque


def main_while_loop(graph, start_node, queue, direction, visited, size):
    neighborhood = {start_node}
    if len(queue) == 0:
        queue.append(start_node)

    while (len(neighborhood) <= size) and len(queue) > 0:
        start = queue.popleft()

        if start not in neighborhood:
            neighborhood.add(start)

        visited.add(start)

        if start == start_node:
            neighbors = graph.nodes[start_node].children(direction)
        else:
            neighbors = graph.nodes[start].neighbors()

        for n in neighbors:
            if n not in visited:
                queue.append(n)

    return neighborhood


def bfs(graph, start_node, size):
    """
    Runs bfs and returns the neighborhood smaller than size

    Using only bfs was resulting in a one-sided neighborhood.
    So the neighborhood I was getting was mainly going from the start node
    into one direction because we have FIFO and it basically keeps going
    in that direction. So I decided to split check if have two possible directions
    From start, too look in both directions separately and add that to the whole neighborhood

    :param graph: A graph object from class Graph
    :param start_node: starting node for the BFS search
    :param size: size of the neighborhood to return
    """

    queue = deque()
    visited = set()

    # I will in both direction, and search for half the neighborhood in one
    # and the other half in the other direction
    # if the nieghborhood is 100, I will traverse 50 to the right and 50 to the left and return that
    queue.append(start_node)
    visited.add(start_node)
    first_direction = graph.nodes[start_node].children(0)
    len_first_direction = 0
    second_direction = graph.nodes[start_node].children(1)
    len_second_direction = 0
    # neighbors = graph.nodes[start_node].neighbors()

    if len(first_direction) == 0:
        if len(second_direction) == 0:
            return {start_node}
        else:
            len_second_direction = size
    else:
        if len(second_direction) == 0:
            len_first_direction = size
        else:
            len_first_direction = int(size/2)
            len_second_direction = int(size/2)

    # if len(neighbors) == 0:  # start node was a lonely node
    #     return neighborhood

    # break when the subgraph is longer than the size wanted
    # or the component the start node in was smaller than the size
    neighborhood = main_while_loop(graph, start_node, queue, 0, visited, len_first_direction)

    queue = deque()
    if len(neighborhood) < int(size/2):
        len_second_direction += size - int(size/2)

    neighborhood = neighborhood.union(main_while_loop(graph, start_node, queue, 1, visited, len_second_direction))

    # while (len(neighborhood) <= len_first_direction) and len(queue) > 0:
    #     start = queue.popleft()
    #
    #     if start not in neighborhood:
    #         neighborhood.add(start)
    #
    #     visited.add(start)
    #
    #     if start == start_node:
    #         neighbors = first_direction
    #     else:
    #         neighbors = graph.nodes[start].neighbors()
    #
    #     for n in neighbors:
    #         if n not in visited:
    #             queue.append(n)
    #
    # queue = deque()
    # queue.append(start_node)
    # # in case on direction had only a few nodes then
    # # then I explore more than half of the size in the other direction
    # if len(neighborhood) < int(size/2):
    #     len_second_direction += size - int(size/2)
    #
    # while (len(neighborhood) <= len_second_direction) and len(queue) > 0:
    #     start = queue.popleft()
    #
    #     if start not in neighborhood:
    #         neighborhood.add(start)
    #
    #     visited.add(start)
    #
    #     if start == start_node:
    #         neighbors = first_direction
    #     else:
    #         neighbors = graph.nodes[start].neighbors()
    #
    #     for n in neighbors:
    #         if n not in visited:
    #             queue.append(n)

    return neighborhood



from BubbleGun.path_extractor import sequence_extractor
import pdb


def output_chains_fasta(graph):

    # Chain ids should be from 1 to len(graph.b_chains)
    chain_id = 1
    hap1_file = open("haplotype1.fasta", "w")
    hap2_file = open("haplotype2.fasta", "w")
    # chain_counter = 1
    for chain in graph.b_chains:
        hap1 = []
        hap2 = []
        if len(chain.sorted) == 0:
            chain.sort()

        # very hacky stuff, don't judge, I wrote it on a train from Dusseldorf to Saarbrucken
        # should do it with topological sorting but I'm too tired
        for b in chain.sorted:
            if len(hap1) == 0:
                # chain_id = b.source.which_chain
                hap1.append(chain.ends[0])
                hap2.append(chain.ends[0])
                hap1.append(b.inside[0].id)
                hap2.append(b.inside[1].id)
                if hap1[len(hap1) - 2] == b.source.id:
                    hap1.append(b.sink.id)
                    hap2.append(b.sink.id)
                else:
                    hap1.append(b.source.id)
                    hap2.append(b.source.id)

            else:
                if hap1[-1] == b.source.id:
                    hap1.append(b.inside[0].id)
                    hap2.append(b.inside[1].id)
                    hap1.append(b.sink.id)
                    hap2.append(b.sink.id)
                else:
                    hap1.append(b.inside[0].id)
                    hap2.append(b.inside[1].id)
                    hap1.append(b.source.id)
                    hap2.append(b.source.id)

        seq1 = sequence_extractor(graph, hap1)
        if len(seq1) == 0:
            pdb.set_trace()
        read_name = ">chain_" + str(chain_id) + "_hap1"
        hap1_file.write(read_name + "\n")
        hap1_file.write(seq1 + "\n")

        seq2 = sequence_extractor(graph, hap2)
        read_name = ">chain_" + str(chain_id) + "_hap2"
        hap2_file.write(read_name + "\n")
        hap2_file.write(seq2 + "\n")
        chain_id += 1

    hap1_file.close()
    hap2_file.close()



#!/usr/bin/env python3
import sys
from distutils.core import setup
from setuptools import setup, find_packages

CURRENT_PYTHON = sys.version_info[:2]
REQUIRED_PYTHON = (3, 3)

# This check and everything above must remain compatible with Python 2.7.
if CURRENT_PYTHON < REQUIRED_PYTHON:
    sys.stderr.write("BubbleGun requires Python 3.3 or higher and "
                     "you current version is {}".format(CURRENT_PYTHON))
    sys.exit(1)


setup(name='BubbleGun',
      version='1.1.7',
      description='Detection of Bubble and Superbubble chains in genome graphs',
      author='Fawaz Dabbaghie',
      author_email='fawaz.dabbaghie@helmholtz-hips.de',
      url='https://github.com/fawaz-dabbaghieh/bubble_gun',
      packages=find_packages(),
      # scripts=['bin/main.py'],
      license="LICENSE.TXT",
      long_description=open("README.md").read(),
      long_description_content_type='text/markdown',
#      install_requires=["protobuf == 3.11.3",
#                        "pystream-protobuf == 1.5.1"],
      # other arguments here...
      entry_points={
          "console_scripts": [
              "BubbleGun=BubbleGun.main:main"
          ]}
      )



import os
import sys
from BubbleGun.Node import Node
import logging


def write_gfa(graph, set_of_nodes=None,
              output_file="output_file.gfa", append=False, optional_info=False):
    """
    Write a gfa out

    :param graph: the graph object
    :param set_of_nodes: A list of node ids of the path or nodes we want to generate a GFA file for.
    :param output_file: path to output file
    :param append: if I want to append to a file instead of rewriting it
    :param optional_info: If set to True, all optional columns for S lines are outputted as well
    :return: writes a gfa file
    """
    nodes = graph.nodes

    if set_of_nodes is None:
        set_of_nodes = graph.nodes.keys()

    if append is False:
        f = open(output_file, "w+")
    else:
        if os.path.exists(output_file):
            f = open(output_file, "a")
        else:
            logging.warning("Trying to append to a non-existent file\n"
                            "creating an output file")
            f = open(output_file, "w+")

    for n1 in set_of_nodes:
        if n1 not in nodes:
            logging.warning("Node {} does not exist in the graph, skipped in output".format(n1))
            continue

        # writing nodes in gfa file
        # if modified:
        #     node = nodes[n1]
        #     # for now I don't care to which sb the node belongs
        #     # I just care about simple bubbles for phasing
        #     specification = str(":".join((str(node.which_chain), str(0),
        #                                   str(node.which_b), str(node.which_allele))))
        #     if nodes[n1].seq == "":
        #         line = str("\t".join(("S", str(n1), str("A" * nodes[n1].seq_len), specification)))
        #     else:
        #         line = str("\t".join(("S", str(n1), nodes[n1].seq, specification)))
        #
        #     f.write(line + "\n")
        # else:
        # if nodes[n1].seq == "":
        #     line = str("\t".join(("S", str(n1), str("A" * nodes[n1].seq_len), "LN:i:" + len())))
        # else:
        line = str("\t".join(("S", str(n1), nodes[n1].seq, "LN:i:" + str(nodes[n1].seq_len))))
        if optional_info:
            line += "\t" + nodes[n1].optional_info

        f.write(line + "\n")

        # writing edges
        edges = []
        # overlap = str(graph.k - 1) + "M\n"

        for n in nodes[n1].start:
            overlap = str(n[2]) + "M\n"
            # I am checking if the are nodes I want to write
            # I think I can remove this later as I implemented the .remove_node
            # to the Graph class that safely removes a node and all its edges
            # So there shouldn't be any edges to removed
            if n[0] in set_of_nodes:
                if n[1] == 0:
                    edge = str("\t".join(("L", str(n1), "-", str(n[0]), "+", overlap)))
                    edges.append(edge)
                else:
                    edge = str("\t".join(("L", str(n1), "-", str(n[0]), "-", overlap)))
                    edges.append(edge)

        for n in nodes[n1].end:
            overlap = str(n[2]) + "M\n"

            if n[0] in set_of_nodes:
                if n[1] == 0:
                    edge = str("\t".join(("L", str(n1), "+", str(n[0]), "+", overlap)))
                    edges.append(edge)
                else:
                    edge = str("\t".join(("L", str(n1), "+", str(n[0]), "-", overlap)))
                    edges.append(edge)

        for e in edges:
            f.write(e)

    f.close()


def write_chains(graph, output_file="output_bubble_chains.gfa", optional_info=False):
    """
    Write bubble chains as gfa file

    :param graph: graph object with bubble chains already detected
    :param output_file: output file path
    :param optional_info: If save memory was used then there are no sequence and this is set to False
    """
    nodes = graph.nodes
    f = open(output_file, "w+")
    set_of_nodes = set()
    # using a set so I do not write the same chains more once
    # in case this chains was a child of another chain
    # I probably can also check the chain's parent sb and parent chains id
    # if it's not 0 then I do not write it, if it is 0 then I write it
    for chain in graph.b_chains:
        for n in chain.list_chain():
            set_of_nodes.add(n)

    for n1 in set_of_nodes:
        # writing nodes in gfa file
        node = nodes[n1]

        if "LN" in nodes[n1].optional_info:
            line = str("\t".join(("S", str(n1), nodes[n1].seq)))
        else:
            line = str("\t".join(("S", str(n1), nodes[n1].seq,  "LN:i:" + str(nodes[n1].seq_len))))
        if optional_info:
            line += "\t" + nodes[n1].optional_info

        f.write(line + "\n")
        # writing edges
        edges = []

        for n in node.start:
            # I check if the neighbor belongs to the same chain
            # otherwise I don't write that edge
            if n[0] in set_of_nodes:
                overlap = str(n[2]) + "M\n"
                if n[1] == 0:
                    edge = str("\t".join(("L", str(n1), "-", str(n[0]), "+", overlap)))
                    edges.append(edge)
                else:
                    edge = str("\t".join(("L", str(n1), "-", str(n[0]), "-", overlap)))
                    edges.append(edge)

        for n in nodes[n1].end:
            overlap = str(n[2]) + "M\n"
            if n[0] in set_of_nodes:
                if n[1] == 0:
                    edge = str("\t".join(("L", str(n1), "+", str(n[0]), "+", overlap)))
                    edges.append(edge)
                else:
                    edge = str("\t".join(("L", str(n1), "+", str(n[0]), "-", overlap)))
                    edges.append(edge)

        for e in edges:
            f.write(e)

    f.close()


def read_gfa(gfa_file_path, low_memory=False):
    """
    Read a gfa file

    :param gfa_file_path: gfa graph file.
    :param low_memory: don't read the sequences to save memory
    :return: Dictionary of node ids and Node objects.
    """
    if not os.path.exists(gfa_file_path):
        logging.error("the gfa file path you gave does not exists, please try again!")
        sys.exit()

    nodes = dict()
    edges = []
    # min_node_length = k
    with open(gfa_file_path, "r") as lines:
        for line in lines:
            if line.startswith("S"):
                line = line.strip().split("\t")
                n_id = str(line[1])
                n_len = len(line[2])
                nodes[n_id] = Node(n_id)

                if len(line) > 3:  # save optional info
                    nodes[n_id].optional_info = "\t".join(line[3:])

                if not low_memory:
                    nodes[n_id].seq_len = n_len
                    nodes[n_id].seq = str(line[2]).strip()

                    # reading extra information about nodes
                    # nodes[n_id].kc = int([x for x in line if x.startswith("KC")][0].split(":")[-1])
                    # nodes[n_id].km = float([x for x in line if x.startswith("km")][0].split(":")[-1])

                    # if min_node_length > nodes[n_id].seq_len:
                    #     logging.error("Node {} has a sequence of length {}"
                    #                   " which is smaller than the provided k\n"
                    #                   "Not allowed.".format(nodes[n_id].id,
                    #                                         nodes[n_id].seq_len))
                    #     sys.exit()

            elif line.startswith("L"):
                edges.append(line)

    for e in edges:
        line = e.split()

        first_node = str(line[1])
        second_node = str(line[3])
        if first_node not in nodes:
            logging.warning(f"an edge between {first_node} and {second_node} exists but a "
                            f"node record for {first_node} does not exist in the file. Skipping")
            continue
        if second_node not in nodes:
            logging.warning(f"an edge between {first_node} and {second_node} exists but a "
                            f"node record for {second_node} does not exist in the file. Skipping")
            continue

        overlap = int(line[5][:-1])

        if line[2] == "-":
            from_start = True
        else:
            from_start = False

        if line[4] == "-":
            to_end = True
        else:
            to_end = False

        if from_start and to_end:  # from start to end L x - y -
            if (second_node, 1, overlap) not in nodes[first_node].start:
                nodes[first_node].start.append((second_node, 1, overlap))
            if (first_node, 0, overlap) not in nodes[second_node].end:
                nodes[second_node].end.append((first_node, 0, overlap))

        elif from_start and not to_end:  # from start to start L x - y +

            if (second_node, 0, overlap) not in nodes[first_node].start:
                nodes[first_node].start.append((second_node, 0, overlap))

            if (first_node, 0, overlap) not in nodes[second_node].start:
                nodes[second_node].start.append((first_node, 0, overlap))

        elif not from_start and not to_end:  # from end to start L x + y +
            if (second_node, 0, overlap) not in nodes[first_node].end:
                nodes[first_node].end.append((second_node, 0, overlap))

            if (first_node, 1, overlap) not in nodes[second_node].start:
                nodes[second_node].start.append((first_node, 1, overlap))

        elif not from_start and to_end:  # from end to end L x + y -
            if (second_node, 1, overlap) not in nodes[first_node].end:
                nodes[first_node].end.append((second_node, 1, overlap))

            if (first_node, 1, overlap) not in nodes[second_node].end:
                nodes[second_node].end.append((first_node, 1, overlap))

    return nodes


# def read_vg(vg_file_path, k):
#     """
#     Read a vg file
#
#     :param vg_file_path: vg graph file.
#     :return: Dictionary of node ids and Node objects.
#     """
#     if not os.path.exists(vg_file_path):
#         logging.error("the vg file path you gave does not exist, please try again!")
#         sys.exit()
#
#     nodes = dict()
#     edges = []
#     min_node_length = k
#     with stream.open(vg_file_path, "rb") as in_stream:
#         for data in in_stream:
#             graph = BubbleGun.vg_pb2.Graph()
#             # import pdb
#             # pdb.set_trace()
#             graph.ParseFromString(data)
#
#             for n in graph.node:
#                 nodes[n.id] = Node(n.id)
#                 nodes[n.id].seq = str(n.sequence)
#                 nodes[n.id].seq_len = len(n.sequence)
#
#                 if min_node_length > nodes[n_id].seq_len:
#                     logging.error("Node {} has a sequence of length {}"
#                                   " which is smaller than the provided k\n"
#                                   "Not allowed.".format(nodes[n_id].id,
#                                                         nodes[n_id].seq_len))
#                     sys.exit()
#
#             for e in graph.edge:
#                 edges.append(e)
#
#     for e in edges:
#         k = getattr(e, "from")
#         neighbor = e.to
#         from_start = e.from_start
#         to_end = e.to_end
#
#         if from_start is True and to_end is True:  # from start to end L x - y -
#             nodes[k].start.append((neighbor, 1))
#             nodes[neighbor].end.append((k, 0))
#
#         elif from_start is True and to_end is False:  # from start to start L x - y +
#             nodes[k].start.append((neighbor, 0))
#             nodes[neighbor].start.append((k, 0))
#
#         elif from_start is False and to_end is False:  # from end to start L x + y +
#             nodes[k].end.append((neighbor, 0))
#             nodes[neighbor].start.append((k, 1))
#
#         elif from_start is False and to_end is True:  # from end to end L x + y -
#             nodes[k].end.append((neighbor, 1))
#             nodes[neighbor].end.append((k, 1))
#
#     return nodes
